/**
  	Each time a DAC interface detects a rising edge on the selected Timer Trigger Output
	(TIMx_TRGO), the last data stored in the DAC_DHRx register is transferred to the
	DAC_DORx register.

	Total of 4APB clock cycles are needed to update the DAC DOR register data. As APB1
	maximum clock is 45 MHz (for STM32F42x), 11.25 Msps is the maximum update rate for the
	DAC output register when timer trigger and the DMA are used for the data update.

	Basic timer (TIM6&TIM7) features include:
	- 16-bit auto-reload upcounter
	- 16-bit programmable prescaler used to divide (also on the fly) the counter clock frequency by any factor between 1 and 65536
	- Synchronization circuit to trigger the DAC
	- Interrupt/DMA generation on the update event: counter overflow
	Counter Register (TIMx_CNT)
	Prescaler Register (TIMx_PSC)
	Auto-Reload Register (TIMx_ARR)
 */

#include "dac.h"
#include "jbuf.h"
#include "stm32f4xx_hal.h"
#include "stm32f4xx_nucleo_144.h"
#include "stm32f4xx_hal_dac.h"
#include "stm32f4xx_hal_tim.h"
#include "asserts.h"
#include "utils.h"
#include <stdio.h>


#define LOCAL_DEBUG
#ifdef LOCAL_DEBUG
#	define TRACE(args) (printf("DAC: "), printf args)
#else
#	define TRACE(args)
#endif
#	define MSG(args) (printf("DAC: "), printf args)


/** \sa htim.Init.Period */
//enum { SAMPLING_FREQUENCY = 44100 };
enum { SAMPLING_FREQUENCY = 16000 };
enum { SAMPLING_FREQUENCY_TUNING = 4 };	///< subtracted from timer period value; not sure why this is needed to match exact frequency
enum { DAC_OFFSET = 32768 };

/* Definition for DAC clock resources */
#define DACx_CHANNEL1_GPIO_CLK_ENABLE()  __HAL_RCC_GPIOA_CLK_ENABLE()
#define DMAx_CLK_ENABLE()               __HAL_RCC_DMA1_CLK_ENABLE()

#define DACx_FORCE_RESET()              __HAL_RCC_DAC_FORCE_RESET()
#define DACx_RELEASE_RESET()            __HAL_RCC_DAC_RELEASE_RESET()

/* Definition for DACx Channel1 Pin */
#define DACx_CHANNEL1_PIN                GPIO_PIN_4
#define DACx_CHANNEL1_GPIO_PORT          GPIOA

/* Definition for DACx's Channel1 */
#define DACx_CHANNEL1                    DAC_CHANNEL_1

/* Definition for DACx's DMA Channel1 */
#define DACx_DMA_CHANNEL1                DMA_CHANNEL_7
#define DACx_DMA_STREAM1                 DMA1_Stream5

/* Definition for DACx's NVIC */
#define DACx_DMA_IRQn1                   DMA1_Stream5_IRQn
#define DACx_DMA_IRQHandler1             DMA1_Stream5_IRQHandler

DAC_HandleTypeDef DacHandle;
static DAC_ChannelConfTypeDef sConfig;
/// \note using DAC_ALIGN_12B_L
/// \note 20 ms = 2 * 10 ms in buffer
uint16_t dacSampleTable[SAMPLING_FREQUENCY/50] = { 0 };
// { 0x0000, 0x1000, 0x2000, 0x3000, 0x4000, 0x5000, 0x6000, 0x7000, 0x8000, 0x9000, 0xA000, 0xB000, 0xC000, 0xD000, 0xE000, 0xF000 };
static void DAC_Ch1_EscalatorConfig(void);
static void TIM6_Config(void);

void dac_init(void) {
	/*##-1- Configure the DAC peripheral #######################################*/
	DacHandle.Instance = DAC;

	/*##-2- Configure the TIM peripheral #######################################*/
	TIM6_Config();

	HAL_DAC_DeInit(&DacHandle);

	DAC_Ch1_EscalatorConfig();
}

static void DAC_Ch1_EscalatorConfig(void) {
	TRACE(("CH1 EscalatorConfig\n"));
	/*##-1- Initialize the DAC peripheral ######################################*/
	if (HAL_DAC_Init(&DacHandle) != HAL_OK) {
		/* Initialization Error */
		Error_Handler();
	}

	/*##-1- DAC channel1 Configuration #########################################*/
	sConfig.DAC_Trigger = DAC_TRIGGER_T6_TRGO;
	sConfig.DAC_OutputBuffer = DAC_OUTPUTBUFFER_ENABLE;

	if (HAL_DAC_ConfigChannel(&DacHandle, &sConfig, DACx_CHANNEL1) != HAL_OK) {
		/* Channel configuration Error */
		Error_Handler();
	}

	/*##-2- Enable DAC Channel1 and associated DMA #############################*/
	TRACE(("Starting DMA\n"));
	if (HAL_DAC_Start_DMA(&DacHandle, DACx_CHANNEL1, (uint32_t*) dacSampleTable,
			ARRAY_SIZE(dacSampleTable), DAC_ALIGN_12B_L) != HAL_OK) {
		/* Start DMA Error */
		TRACE(("Error starting DMA\n"));
		Error_Handler();
	}
	TRACE(("DMA started\n"));
}

/**
 * @brief  TIM6 Configuration
 * @note   TIM6 configuration is based on APB1 frequency
 * @note   TIM6 Update event occurs each TIM6CLK/256
 */
void TIM6_Config(void) {
	static TIM_HandleTypeDef htim;
	TIM_MasterConfigTypeDef sMasterConfig;

	/*##-1- Configure the TIM peripheral #######################################*/
	/* Time base configuration */
	htim.Instance = TIM6;

	htim.Init.Period = 90000000/SAMPLING_FREQUENCY - SAMPLING_FREQUENCY_TUNING; //0x7F4; //roughly 44.1kHz (90000000/2047);
	htim.Init.Prescaler = 0;
	htim.Init.ClockDivision = 0;
	// The counter counts from 0 to the auto-reload value (contents of the TIMx_ARR register), then restarts from 0 and generates a counter overflow event.
	htim.Init.CounterMode = TIM_COUNTERMODE_UP;
	HAL_TIM_Base_Init(&htim);

	/* TIM6 TRGO selection */
	sMasterConfig.MasterOutputTrigger = TIM_TRGO_UPDATE;
	sMasterConfig.MasterSlaveMode = TIM_MASTERSLAVEMODE_DISABLE;

	HAL_TIMEx_MasterConfigSynchronization(&htim, &sMasterConfig);

	/*##-2- Enable TIM peripheral counter ######################################*/
	HAL_TIM_Base_Start(&htim);
}

void HAL_DAC_MspInit(DAC_HandleTypeDef* hdac) {
	GPIO_InitTypeDef GPIO_InitStruct;
	static DMA_HandleTypeDef hdma_dac1;

	/*##-1- Enable peripherals and GPIO Clocks #################################*/
	/* DAC Periph clock enable */
	__HAL_RCC_DAC_CLK_ENABLE()
	;
	/* Enable GPIO clock ****************************************/
	DACx_CHANNEL1_GPIO_CLK_ENABLE()
	;
	/* DMA1 clock enable */
	DMAx_CLK_ENABLE()
	;

	/*##-2- Configure peripheral GPIO ##########################################*/
	/* DAC Channel1 GPIO pin configuration */
	GPIO_InitStruct.Pin = DACx_CHANNEL1_PIN;
	GPIO_InitStruct.Mode = GPIO_MODE_ANALOG;
	GPIO_InitStruct.Pull = GPIO_NOPULL;
	HAL_GPIO_Init(DACx_CHANNEL1_GPIO_PORT, &GPIO_InitStruct);

	/*##-3- Configure the DMA streams ##########################################*/
	/* Set the parameters to be configured for Channel1*/
	hdma_dac1.Instance = DACx_DMA_STREAM1;

	hdma_dac1.Init.Channel = DACx_DMA_CHANNEL1;
	hdma_dac1.Init.Direction = DMA_MEMORY_TO_PERIPH;
	hdma_dac1.Init.PeriphInc = DMA_PINC_DISABLE;
	hdma_dac1.Init.MemInc = DMA_MINC_ENABLE;
	hdma_dac1.Init.PeriphDataAlignment = DMA_PDATAALIGN_HALFWORD;
	hdma_dac1.Init.MemDataAlignment = DMA_MDATAALIGN_HALFWORD;
	hdma_dac1.Init.Mode = DMA_CIRCULAR;
	hdma_dac1.Init.Priority = DMA_PRIORITY_HIGH;
	hdma_dac1.Init.FIFOMode = DMA_FIFOMODE_DISABLE;
	hdma_dac1.Init.FIFOThreshold = DMA_FIFO_THRESHOLD_HALFFULL;
	hdma_dac1.Init.MemBurst = DMA_MBURST_SINGLE;
	hdma_dac1.Init.PeriphBurst = DMA_PBURST_SINGLE;

	HAL_DMA_Init(&hdma_dac1);

	/* Associate the initialized DMA handle to the the DAC handle */
	__HAL_LINKDMA(hdac, DMA_Handle1, hdma_dac1);

	/*##-4- Configure the NVIC for DMA #########################################*/
	/* Enable the DMA1 Stream5 IRQ Channel */
	HAL_NVIC_SetPriority(DACx_DMA_IRQn1, 2, 0);
	HAL_NVIC_EnableIRQ(DACx_DMA_IRQn1);
}

/**
  * @brief Tx Transfer completed callbacks.
  */
void HAL_DAC_ConvCpltCallbackCh1(DAC_HandleTypeDef* hdac)
{
	//TRACE(("F\n"));
	uint16_t *ptr = dacSampleTable + ARRAY_SIZE(dacSampleTable)/2;
	short *src = jbuf_get();
	for (int i=0; i<ARRAY_SIZE(dacSampleTable)/2; i++) {
		ptr[i] = src[i] + DAC_OFFSET;
	}
}

/**
  * @brief Tx Transfer Half completed callbacks
  */
void HAL_DAC_ConvHalfCpltCallbackCh1(DAC_HandleTypeDef* hdac)
{
	//TRACE(("H\n"));
	uint16_t *ptr = dacSampleTable;
	short *src = jbuf_get();
	for (int i=0; i<ARRAY_SIZE(dacSampleTable)/2; i++) {
		ptr[i] = src[i] + DAC_OFFSET;
	}
}

void HAL_DAC_ErrorCallbackCh1(DAC_HandleTypeDef *hdac) {
	MSG(("ErrorCallbackCh1!\n"));
}

void HAL_DAC_DMAUnderrunCallbackCh1(DAC_HandleTypeDef *hdac) {
	MSG(("UnderrunCallbackCh1!\n"));
}

/**
 * @brief  DeInitializes the DAC MSP.
 * @param  hdac: pointer to a DAC_HandleTypeDef structure that contains
 *         the configuration information for the specified DAC.
 */
void HAL_DAC_MspDeInit(DAC_HandleTypeDef* hdac) {
	static DMA_HandleTypeDef hdma_dac1;

	/*##-1- Reset peripherals ##################################################*/
	DACx_FORCE_RESET();
	DACx_RELEASE_RESET();

	/*##-2- Disable peripherals and GPIO Clocks ################################*/
	/* De-initialize the DAC Channel1 GPIO pin */
	HAL_GPIO_DeInit(DACx_CHANNEL1_GPIO_PORT, DACx_CHANNEL1_PIN);

	/*##-3- Disable the DMA Streams ############################################*/
	/* De-Initialize the DMA Stream associate to Channel1 */
	hdma_dac1.Instance = DACx_DMA_STREAM1;
	HAL_DMA_DeInit(&hdma_dac1);

	/*##-4- Disable the NVIC for DMA ###########################################*/
	HAL_NVIC_DisableIRQ(DACx_DMA_IRQn1);
}

/**
 * @brief TIM MSP Initialization
 *        This function configures the hardware resources used in this example:
 *           - Peripheral's clock enable
 *           - Peripheral's GPIO Configuration
 * @param htim: TIM handle pointer
 */
void HAL_TIM_Base_MspInit(TIM_HandleTypeDef* htim) {
	/* TIM6 Periph clock enable */
	__HAL_RCC_TIM6_CLK_ENABLE()
	;
}

/**
 * @brief TIM MSP De-Initialization
 *        This function frees the hardware resources used in this example:
 *          - Disable the Peripheral's clock
 *          - Revert GPIO to their default state
 * @param htim: TIM handle pointer
 */
void HAL_TIM_Base_MspDeInit(TIM_HandleTypeDef *htim) {

	/*##-1- Reset peripherals ##################################################*/
	__HAL_RCC_TIM6_FORCE_RESET();
	__HAL_RCC_TIM6_RELEASE_RESET();
}

/******************************************************************************/
/*                 STM32F4xx Peripherals Interrupt Handlers                   */
/*  Add here the Interrupt Handler for the used peripheral(s) (DAC), for the  */
/*  available peripheral interrupt handler's name please refer to the startup */
/*  file (startup_stm32f4xx.s).                                               */
/******************************************************************************/

/**
 * @brief  This function handles DMA interrupt request.
 * @param  None
 * @retval None
 */
void DACx_DMA_IRQHandler1(void) {
	HAL_DMA_IRQHandler(DacHandle.DMA_Handle1);
}

/**
 * @brief  This function handles DMA interrupt request.
 * @param  None
 * @retval None
 */
void DACx_DMA_IRQHandler2(void) {
	HAL_DMA_IRQHandler(DacHandle.DMA_Handle2);
}

